# Copyright (c)  2020  Mobvoi Inc.        (authors: Fangjun Kuang)
#                      Guoguo Chen
#
# See ../../../LICENSE for clarification regarding multiple authors

from collections import OrderedDict
from typing import Optional
from typing import Union

import torch

from .symbol_table import SymbolTable
from _k2 import RaggedArc
from _k2 import _Fsa
from _k2 import _as_float
from _k2 import _fsa_from_str
from _k2 import _fsa_from_tensor
from _k2 import _fsa_to_str
from graphviz import Digraph


class Fsa(object):
    '''This class represents a single fsa or a vector of fsas.

    When it denotes a single fsa, its attribute :attr:`shape` is a tuple
    containing two elements ``(num_states, None)``; it is a tuple with three
    elements ``(num_fsas, None, None)`` for a vector of fsas.
    '''

    def __init__(self,
                 tensor: torch.Tensor,
                 aux_labels: Optional[torch.Tensor] = None) -> 'Fsa':
        '''Build an Fsa from a tensor with optional aux_labels.

        It is useful when loading an Fsa from file.

        Args:
          tensor:
            A torch tensor of dtype `torch.int32` with 4 columns.
            Each row represents an arc. Column 0 is the src_state,
            column 1 the dest_state, column 2 the label, and column
            3 the score.

            Caution:
              Scores are floats and their binary pattern is
              **reinterpreted** as integers and saved in a tensor
              of dtype `torch.int32`.

          aux_labels:
            Optional. If not None, it associates an aux_label with every arc,
            so it has as many rows as `tensor`. It is a 1-D tensor of dtype
            `torch.int32`.

        Returns:
          An instance of Fsa.
        '''
        self._arcs = _fsa_from_tensor(tensor)
        self._attr_dict = OrderedDict()
        self._aux_labels = aux_labels

    @classmethod
    def _create(cls, fsa: _Fsa,
                aux_labels: Optional[torch.Tensor] = None) -> 'Fsa':
        '''For internal use only.
        '''
        ans = cls.__new__(cls)
        super(Fsa, ans).__init__()
        ans._fsa = fsa
        ans._aux_labels = aux_labels
        return ans

    @classmethod
    def from_str(cls, s: str):
        '''Create an Fsa from a string.

        The given string `s` consists of lines with the following format:

        (1) When it represents an acceptor:

                src_state dest_state label score

        (2) When it represents a transducer:

                src_state dest_state label aux_label score

        The line for the final state consists of only one field:

                final_state

        Note:
          Fields are separated by space(s), tab(s) or both. The `score`
          field is a float, while other fields are integers.

        Caution:
          The first column has to be non-decreasing.

        Caution:
          The final state has the largest state number. There is only
          one final state. All arcs that are connected to the final state
          have label -1.

        Args:
          s:
            The input string. Refer to the above comment for its format.
        '''
        # Figure out acceptor/transducer for k2 fsa.
        acceptor = True
        line = s.strip().split('\n', 1)[0]
        fields = line.strip().split()
        assert len(fields) == 4 or len(fields) == 5
        if len(fields) == 5:
            acceptor = False

        ans = cls.__new__(cls)
        super(Fsa, ans).__init__()
        fsa, aux_labels = _fsa_from_str(s, acceptor, False)
        ans._fsa = fsa
        ans._aux_labels = aux_labels
        return ans

    @classmethod
    def from_openfst(cls, s: str, acceptor: bool = True) -> 'Fsa':
        '''Create an Fsa from a string in OpenFST format.

        The given string `s` consists of lines with the following format:

        (1) When it represents an acceptor:

                src_state dest_state label score

        (2) When it represents a transducer:

                src_state dest_state label aux_label score

        The line for the final state consists of two fields:

                final_state score

        Note:
          Fields are separated by space(s), tab(s) or both. The `score`
          field is a float, while other fields are integers.

          There might be multiple final states. Also, OpenFST may omit the score
          if it is 0.0.

        Args:
          s:
            The input string. Refer to the above comment for its format.
          acceptor:
            Optional. If true, interpret the input string as an acceptor,
            otherwise, interpret it as a transducer.
        '''
        ans = cls.__new__(cls)
        super(Fsa, ans).__init__()
        fsa, aux_labels = _fsa_from_str(s, acceptor, True)
        ans._fsa = fsa
        ans._aux_labels = aux_labels
        return ans

    def set_isymbol(self, isym: SymbolTable) -> None:
        '''Set the input symbol table.

        Args:
          isym:
            The input symbol table.
        Returns:
          None.
        '''
        self.isym = isym

    def set_osymbol(self, osym: SymbolTable) -> None:
        '''Set the output symbol table.

        Args:
          osym:
            The output symbol table.

        Returns:
          None.
        '''
        self.osym = osym

    def to_str(self, openfst: bool = False) -> str:
        '''Convert an Fsa to a string.

        Note:
          The returned string can be used to construct an Fsa.

        Args:
          openfst:
            Optional. If true, we negate the score during the conversion,

        Returns:
          A string representation of the Fsa.
        '''
        return _fsa_to_str(self._fsa, openfst, self._aux_labels)

    @property
    def arcs(self) -> RaggedArc:
        return self._arcs

    @property
    def weights(self) -> torch.Tensor:
        '''Returns the weights of arcs in the Fsa.

        Returns:
          A 1-D tensor of dtype `torch.float32`. It has
          as many rows as `arcs`.
        '''
        return _as_float(self._fsa.values.tensor()[:, -1])

    @property
    def aux_labels(self) -> Union[torch.Tensor, None]:
        '''Return the aux_labels associated with `arcs`, if any.

        Returns:
          None or a 1-D tensor of dtype `torch.int32` if the Fsa
          is a transducer. It has as many rows as `arcs`.
        '''
        return self._aux_labels

    def to(self, device: Union[str, torch.device]) -> 'Fsa':
        '''Convert an Fsa to a new Fsa on a given device.

        Caution:
          It is NOT an in-place operation. It returns a NEW instance.

        Args:
          device:
            A torch device. Currently it supports only CUDA and CPU devices.

        Returns:
          A new Fsa on the given device.
        '''
        if isinstance(device, str):
            device = torch.device(device)
        assert device.type in ['cpu', 'cuda']

        if device.type == 'cuda':
            fsa = self._fsa.cuda(device.index if device.index else -1)
        else:
            fsa = self._fsa.cpu()

        if self._aux_labels is not None:
            aux_labels = self._aux_labels.to(device)
        else:
            aux_labels = None
        return Fsa._create(fsa, aux_labels)

    def to_dot(self) -> Digraph:
        if self._aux_labels is not None:
            name = 'WFST'
        else:
            name = 'WFSA'
        graph_attr = {
            'rankdir': 'LR',
            'size': '8.5,11',
            'label': '',
            'center': '1',
            'orientation': 'Portrait',
            'ranksep': '0.4',
            'nodesep': '0.25',
        }

        default_node_attr = {
            'shape': 'circle',
            'style': 'bold',
            'fontsize': '14',
        }

        final_state_attr = {
            'shape': 'doublecircle',
            'style': 'bold',
            'fontsize': '14',
        }

        final_state = -1
        dot = Digraph(name=name, graph_attr=graph_attr)

        seen = set()
        i = -1
        for arc, weight in zip(self.arcs, self.weights.tolist()):
            i += 1
            src_state, dst_state, label = arc.tolist()
            src_state = str(src_state)
            dst_state = str(dst_state)
            label = int(label)
            if label == -1:
                final_state = dst_state
            if src_state not in seen:
                dot.node(src_state, label=src_state, **default_node_attr)
                seen.add(src_state)

            if dst_state not in seen:
                if dst_state == final_state:
                    dot.node(dst_state, label=dst_state, **final_state_attr)

                else:
                    dot.node(dst_state, label=dst_state, **default_node_attr)
                seen.add(dst_state)
            if self._aux_labels is not None:
                aux_label = int(self._aux_labels[i])
                if hasattr(self, 'osym'):
                    aux_label = self.osym.get(aux_label)
                aux_label = f':{aux_label}'
            else:
                aux_label = ''

            if hasattr(self, 'isym') and label != -1:
                label = self.isym.get(label)

            dot.edge(src_state,
                     dst_state,
                     label=f'{label}{aux_label}/{weight:.2f}')
        return dot
